#include "serverLoop.h"
#include "RakNet\RakSleep.h"

#include "Common.h"
#include "User.h"
#include "legoPackets.h"
#include <iomanip>
#include <sstream>
#include <mutex>
#include <fstream>

#include "Raknet\RSACrypt.h"

using namespace std;

uint packetCount = 0;
std::mutex consoleoutmutex;

void InitSecurity(RakPeerInterface *rakServer, bool useEncryption) {
	// luni client requires this password, seems to be the same for all versions (even alpha/beta)
	rakServer->SetIncomingPassword("3.25 ND1", 8);
	if (!useEncryption) return;

	// These are the sizes necessary for e,n,p,q
	big::u32 e;	RSA_BIT_SIZE n;									// e,n is the public key
	BIGHALFSIZE(RSA_BIT_SIZE, p); BIGHALFSIZE(RSA_BIT_SIZE, q);	// p,q is the private key

	bool keyLoaded = false;
	FILE* fp = fopen("public_key.bin", "rb");
	if (fp) {
		printf("Loading public key... ");
		fread((char*)(&e), sizeof(e), 1, fp);
		fread((char*)(n), sizeof(n), 1, fp);
		fclose(fp);
	} else
		printf("Failed to open \"public_key.bin\".\n");

	fp = fopen("private_key.bin", "rb");
	if (fp) {
		printf("Loading private key... ");
		fread(p, sizeof(RSA_BIT_SIZE) / 2, 1, fp);
		fread(q, sizeof(RSA_BIT_SIZE) / 2, 1, fp);
		fclose(fp);
		printf("Done.\n");
		keyLoaded = true;
	} else
		printf("Failed to open \"private_key.bin\".\n");

	big::RSACrypt<RSA_BIT_SIZE> rsacrypt;
	if (!keyLoaded)		// if there's no predefined key just generate one (and write them on disk)
	{
		printf("Generating %i byte key.\n", sizeof(RSA_BIT_SIZE));

		rsacrypt.generateKeys();
		rsacrypt.getPublicKey(e, n);
		rsacrypt.getPrivateKey(p, q);
		printf("Keys generated.\n");
		printf("Writing public key... ");
		FILE* fp = fopen("public_key.bin", "wb");
		fwrite((char*)&e, sizeof(e), 1, fp);
		fwrite((char*)n, sizeof(n), 1, fp);
		fclose(fp);

		printf("Writing private key... ");
		fp = fopen("private_key.bin", "wb");
		fwrite(p, sizeof(RSA_BIT_SIZE) / 2, 1, fp);
		fwrite(q, sizeof(RSA_BIT_SIZE) / 2, 1, fp);
		fclose(fp);
		printf("Done.\n");
	} else
		rsacrypt.setPrivateKey(p, q);

	//rakServer->InitializeSecurity(0,0,0,0);					// directly generate the keys
	rakServer->InitializeSecurity(0, 0, (char*)p, (char*)q);	// e and n are actually not required
}

string RawDataToString(uchar* data, uint size) {
	ostringstream buffer;
	buffer << "data in bytes:";
	for (uint i = 0; i < size; i++) {
		if (i % 16 == 0) buffer << "\n\t\t";
		else buffer << " ";
		buffer << setw(2) << hex << setfill('0') << (int)data[i];
	}
	buffer << "\n\n";
	return buffer.str();
}

void PrintPacketInfo(Packet *packet, PacketFileLogger* msgFileHandler) {
	stringstream buffer;
	bool identified = true;

	switch (packet->data[0]) {
		// todo: treat every case (excluding internal MsgID's since they don't go here) that occurs in traffic here so the user (not end-user but a dev) knows whats going on
		case ID_RSA_PUBLIC_KEY_MISMATCH:
			buffer << "Public key mismatch.\nThe connecting system's public key does not\nmatch what the sender sent.\n";
			break;
		case ID_CONNECTION_REQUEST_ACCEPTED:
			buffer << "Connection request accepted.\n";
			break;
		case ID_NEW_INCOMING_CONNECTION:
			buffer << "New incoming connection.\n";
			break;
		case ID_MODIFIED_PACKET:
			buffer << "Packet checksum invalid.  Either RSA decrypt function gave the wrong value\nor the packet was tampered with.\n";
			break;
		case ID_DISCONNECTION_NOTIFICATION:
			buffer << "Client disconnected.\n";
			packetCount = 0;
			break;
		case ID_CONNECTION_LOST:
			buffer << "Connection lost.\n";
			packetCount = 0;
		default:
			identified = false;
			if (packet->data[3] != 5 && packet->data[3] != 22)	// dont spam disk with position updates from the client (only show in console/log)
			{
				packetCount++;
				// save the packets that the client sent to us so every session can easily be analyzed
				
				buffer << ".\\clientResponses\\clientToServer" << packetCount << ".bin";
				FILE* fp = fopen( buffer.str().c_str(), "wb" );
				if (fp) {
					fwrite(packet->data, packet->length, 1, fp);
					fclose(fp);
				}
			}

			if (packet->data[0] == ID_LEGO_PACKET &&
				(packet->data[1] == 0 || packet->data[1] == 1 ||
				(packet->data[1] == 4 &&
				(packet->data[3] == 1 || packet->data[3] == 2 || packet->data[3] == 3 || packet->data[3] == 4 || packet->data[3] == 19)))
				)
				buffer << "received \"known\" packet, header: " << packet->data[1] << ", subHeader: " << packet->data[3] << "length : " << packet->length << + "\n";
			else	// note: if the packet ever contains a big readable string this could lead to a potentional buffer overflow (but that case is unlikely)
				buffer << "Unknown ID: " << packet->data[0] << " header: " << packet->data[1] << ", subHeader : " << packet->data[3] << ", length : " << packet->length << ", bitlength : " << packet->bitSize << ", data : " << packet->data << "\n";
	}

	if (msgFileHandler != NULL)	// check if log is disabled
		msgFileHandler->WriteLog( buffer.str().c_str() );

	cout << buffer.str();
	buffer.clear();
}

vector<uchar> OpenPacket(const string& filename) {
	ifstream file(filename, ios::in | ios::binary | ios::ate);
	if (file.is_open()) {
		int fsz = file.tellg();
		vector<uchar> r( fsz );
		file.seekg(0, ios::beg);

		file.read((char*)r.data(), fsz);

		file.close();

		return r;
	}
	consoleoutmutex.lock();
	cout << "Failed to open: " << filename << endl;
	consoleoutmutex.unlock();
	return vector<uchar>(0);
}

void listenForPackets(CONNECT_INFO* cfg, const string& servernam) {
	RakPeerInterface* rakServer = RakNetworkFactory::GetRakPeerInterface();
	
	PacketFileLogger* msgFileHandler = NULL;
	if (cfg->logFile) {
		msgFileHandler = new PacketFileLogger();
		rakServer->AttachPlugin(msgFileHandler);
	}

	InitSecurity(rakServer, cfg->useEncryption);
	SocketDescriptor socketDescriptor(cfg->listenPort, 0);	

	if ( rakServer->Startup(8, 30, &socketDescriptor, 1) ) {
		consoleoutmutex.lock();
		cout << servernam << " started! Listening on: " << cfg->listenPort << "\n";
		consoleoutmutex.unlock();
	} else exit(2);
	rakServer->SetMaximumIncomingConnections(8);

	if (msgFileHandler != NULL) msgFileHandler->StartLog(servernam.c_str());

	Packet* packet;

	while (!LUNIterminate)		// todo: make loop escapable so the program can exit properly
	{
		RakSleep(30);	// This sleep keeps RakNet responsive
		packet = rakServer->Receive();
		if (packet) {
			PrintPacketInfo(packet, msgFileHandler);
			// create and send packets back here according to the one we got
			if (packet->data[0] == ID_LEGO_PACKET) {
				// NOTE: I only worked the ID's up to the character selection so there might be some conflicts with world traffic
				// as long as we don't know the exact packet contents (so we can build it ourselves) we have to send the raw data from the captured traffics
				if (packet->data[1] == 0)	// seems to be always there (from client+server) when initiating a connection (for version check, handshake, etc?)
				{
					if (packet->data[3] == 0)	// thats just a formality so far since no other numbers occured for this case
					{
						// response: 00, 00
						auto v = OpenPacket(".\\" + servernam + "\\init_aw.bin");

						if (v.size() > 0) {
							// bytes 8-10 is the game version (same numbers are mentioned in the game log)
							// the client expects it to be the same that he sent (otherwise he displays "server is being updated" and disconnects)
							rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);
						}
					}
				} else if (packet->data[1] == 1)	// contains user login info from client?
				{
					if (packet->data[3] == 0)	// thats just a formality so far since no other numbers occured for this case
					{
						// response: 05, 00
						string username, password;
						// get the username just for fun
						if (packet->length > 0x8)	// this is where username always starts (how many bytes are reserved for it in the packet?)
						{
							int i = 0;
							while (packet->data[0x8 + i] != '\0' && packet->length > (uint)(0x8 + i)) {
								username.push_back(packet->data[8 + i]);
								i += 2;
							}
							if (username.empty()) {
								consoleoutmutex.lock();
								cout << "Couldn't read username!\n";
								consoleoutmutex.unlock();
							}
						}
						// get the password just for fun
						if (packet->length > 0x4a)	// this is where password starts (how many bytes are reserved for it in the packet?)
						{
							int i = 0;
							while (packet->data[0x4a + i] != '\0' && packet->length > (uint)(0x4a + i)) {
								password.push_back(packet->data[0x4a + i]);
								i += 2;
							}
							if (username.empty()) {
								consoleoutmutex.lock();
								cout << "Couldn't read password!\n";
								consoleoutmutex.unlock();
							}
						}
						printf("User %s connected with password %s\n", username.c_str(), password.c_str());

						unique_ptr<User> usr = User::Login(username, password);
						// we could do a login check here and then send the according response but as long as we use predefined data to send away its not worth the trouble
						/*if( usr == NULL )	// very simple check
						{
							consoleoutmutex.lock();
							cout << "Wrong user information\n";
							consoleoutmutex.unlock();
							auto v = OpenPacket(".\\" + servernam + "\\auth_aw_login_bad.bin");
							if(v.size() > 0) {
								rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);
							}
						}
						else {*/
							// 0x159 is where IP to redirect to starts (don't know how much bytes are reserved for it in the packet, I would imagine 16
							// but since there was always a great zero gap until the port in all these captured packets its nothing to worry about (yet)
							auto v = OpenPacket(".\\" + servernam + "\\auth_aw_login_ok.bin");
							if (v.size() > 0) {
								if (v.size() > 0x19B)	// 0x19B is port (2 bytes reserved or more?)
								{
									if (cfg->redirectIp[0] != 0)
										memcpy(v.data() + 0x159, cfg->redirectIp, sizeof(cfg->redirectIp));
									if (cfg->redirectPort > 0)
										memcpy(v.data() + 0x19B, &cfg->redirectPort, sizeof(cfg->redirectPort));
								}
								rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);
							}
						//}
					}
				} else if (packet->data[1] == 4)	// all packets that have this ID came from the client so far
				{
					if (packet->data[3] == 2)		// this always comes together with id 1 during user login so respond to this id in character traffic (once we figured out how to check if traffic is world or char)
					{
						// do nothing until we know how if the ID=1 packet came from the character selection or world traffic (is that even possible only with the packet data?)
					} else if (packet->data[3] == 1)	// this comes both in user login and connect world though in the latter it comes alone and the answer is different, so how to check for that?
					{
						// char response: 05, 06
						// world response: 05, 02
						// char traffic only - byte 8 defines number of characters, if there are no characters the user automatically gets to the character creation screen
						auto v = OpenPacket(".\\" + servernam + "\\char_aw2.bin");
						if (v.size() > 0) {
							rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);
						}
					} else if (packet->data[3] == 3)	// user created new minifig so we have to analyze packet and send updated character info? (if everything goes well we'll receive id 4 after this)
					{
						// response: 05, 06
						// since we can't build our own packets yet we can't send a response here (because it is dependent on the user input)
					} else if (packet->data[3] == 4)	// is this sent from the client once the user wants to enter world?
					{
						// response: 05, 14
						auto v = OpenPacket(".\\" + servernam + "\\char_aw_redirect.bin");
						if (v.size() > 0) {
							// ip starts at byte 8, port starts at byte 0x29
							if (cfg->redirectIp[0] != 0)
								memcpy(v.data() + 0x8, cfg->redirectIp, sizeof(cfg->redirectIp) - 3);	// NOTE: the IP can currently only be 13 characters long here since there are some non-zero bytes directly after the IP and I don't want to delete them (is it safe to overwrite them?)
							if (cfg->redirectPort > 0)
								memcpy(v.data() + 0x29, &cfg->redirectPort, sizeof(cfg->redirectPort));
							rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);
						}
						rakServer->CloseConnection(packet->systemAddress, true);
					} else if (packet->data[3] == 6)	// user deleted minifig
					{
						// do nothing since we are not storing user information yet (and have no captured packets here)
					}
					else if (packet->data[3] == 19)      // comes from client 3rd response (and _not_ bundled with id 1) from world
					{
						// response: 05, 04
						// the answer to this seems to be pretty big, no idea what data it could be though (maybe progress stuff for the character, missions and backpack? or information about people on the server?)
						// also since the answer seems to consist of a split packet, does RakNet automatically split the packets if the data is too big? or do we have to do something, if yes what?
						// byte 08 (2 or 4 bytes long? I guess 4...) seems to define the size of the data we send, no idea what the data itself is though...

						auto v = OpenPacket(".\\" + servernam + "\\rocket_1a.bin");
						if (v.size() > 0) {
							rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);

						}

						v = OpenPacket(".\\" + servernam + "\\rocket_1b.bin");
						if (v.size() > 0) {
							rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);

						}

						v = OpenPacket(".\\" + servernam + "\\rocket_1c.bin");
						if (v.size() > 0) {
							rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);

						}

						for (int i = 0; i < 5; i++)
						{
							stringstream fl;
							fl << ".\\" << servernam << "\\loading_1_" << (i + 1) << ".bin";
							auto v = OpenPacket(fl.str());
							if (v.size() > 0) {
								rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);
							}
						}

						v = OpenPacket(".\\" + servernam + "\\world_2a.bin");
						if (v.size() > 0) {
							rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);

						}
						// test, is this the right thing to do? there's no difference in the process when they are (not) sent so far
						v = OpenPacket(".\\" + servernam + "\\world_2b.bin");   // I assume these two are sent together?
						if (v.size() > 0) {
							rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);

						}

						for (int i = 0; i < 42; i++)   // herpaderp
						{
							stringstream fl;
							fl << ".\\" << servernam << "\\test\\world_2_" << (i + 1) << ".bin";
							auto v = OpenPacket(fl.str());
							if (v.size() > 0) {
								rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);

							}
						}

						for (int i = 0; i < 1; i++)   // herpaderp
						{
							stringstream fl;
							fl << ".\\" << servernam << "\\test\\world_2_.bin";
							auto v = OpenPacket(fl.str());
							if (v.size() > 0) {
								rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);

							}
						}

						for (int i = 42; i < 210; i++)   // herpaderp
						{
							stringstream fl;
							fl << ".\\" << servernam << "\\test\\world_2_" << (i + 1) << ".bin";
							auto v = OpenPacket(fl.str());
							if (v.size() > 0) {
								rakServer->Send((const char*)v.data(), v.size(), SYSTEM_PRIORITY, RELIABLE_ORDERED, 0, packet->systemAddress, false);

							}
						}
					}
					// these are the other packets from a short world traffic
					else if (packet->data[3] == 21 ||	// (0x15) shortly after init, 3 packets
						packet->data[3] == 5 ||		// (0x05) some occurrences at the beginning (and the end?) of the short traffic
						packet->data[3] == 22 ||	// (0x16) also quite a few of these but only starting in the latter half of the traffic (probably status update from client?)
						packet->data[3] == 23)		// (0x17) only one packet in the traffic with zeroes afterwards, no idea
					{
						// do nothing as long as we don't know what to respond
					}
					// else do nothing, other cases didn't occur so far (didn't check any longer world traffic yet)
				} else if (packet->data[1] == 5)		// all packets that have this ID came from the server so far
				{
					// do nothing since we are supposed to send these packets, not receive them
					// packet->data[3] == 0		// 2nd response auth (if user info is valid), contains redirection to char?
					// packet->data[3] == 2		// 2nd response world, ???
					// packet->data[3] == 4		// 3rd response from world, ??? (contains big data)
					// packet->data[3] == 6		// 2nd response char, contains (all) character data?
					// packet->data[3] == 7		// occured after user created minfig, along with 6
					// packet->data[3] == 14	// 3rd response char, contains redirection to world?
					// don't know yet when we have to use these:
					// packet->data[3]  == 62 (0x3E)// directly after big one from server, only one occurrence in that traffic
					// packet->data[3]  == 12 (0x0C)// lots of this one from the server, seem to describe properties of objects?
					// packet->data[3]  == 49 (0x31)// somewhere during traffic, one occurrence in that traffic
					// packet->data[3]  == 30 (0x1E)// shortly after 49, one occurrence in that traffic
					// packet->data[3]  == 34 (0x22)// directly after 30, one occurrence in that traffic
				}
				// else do nothing, other cases didn't occur so far
			}
			rakServer->DeallocatePacket(packet);
		}
	}

	consoleoutmutex.lock();
	cout << "Quitting " << servernam << endl;
	consoleoutmutex.unlock();
	rakServer->Shutdown(0);
	RakNetworkFactory::DestroyRakPeerInterface(rakServer);
}